import pytest
import json
import tempfile
from pathlib import Path
from textwrap import dedent

import skylos
from skylos.analyzer import Skylos
from skylos.constants import DEFAULT_EXCLUDE_FOLDERS


class TestSkylosIntegration:
    @pytest.fixture
    def temp_project(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            project_path = Path(temp_dir)

            pyproject = project_path / "pyproject.toml"
            pyproject.write_text("[tool.skylos]\n")

            main_py = project_path / "main.py"
            main_py.write_text(
                dedent("""
                import os  # unused import
                import sys
                from typing import List  # unused import
                from collections import defaultdict
                from utils import UsedClass, exported_function
                
                def used_function(x):
                    '''This function is called and should not be flagged'''
                    return x * 2
                
                def unused_function(a, b):
                    '''This function is never called'''
                    unused_var = a + b  # unused variable
                    return unused_var
                
                def main():
                    result = used_function(5)
                    print(result, file=sys.stderr)
                    data = defaultdict(list)
                    
                    obj = UsedClass()
                    obj.get_value()
                    exported_function()
                    
                    return data
                
                if __name__ == "__main__":
                    main()
            """)
            )

            utils_py = project_path / "utils.py"
            utils_py.write_text(
                dedent("""
                import json  # unused import
                
                class UsedClass:
                    '''This class is imported and used'''
                    def __init__(self):
                        self.value = 42
                    
                    def get_value(self):
                        return self.value
                
                class UnusedClass:
                    '''This class is never used'''
                    def __init__(self):
                        self.data = {}
                    
                    def unused_method(self, param):  # unused parameter
                        return "never called"
                
                def exported_function():
                    '''This function is used by main.py'''
                    return "exported"
            """)
            )

            package_dir = project_path / "mypackage"
            package_dir.mkdir()

            init_py = package_dir / "__init__.py"
            init_py.write_text(
                dedent("""
                from .submodule import PublicClass
                
                __all__ = ['PublicClass', 'public_function']
                
                def public_function():
                    return "public"
                
                def _private_function():
                    return "private"
            """)
            )

            sub_py = package_dir / "submodule.py"
            sub_py.write_text(
                dedent("""
                class PublicClass:
                    '''Exported via __init__.py'''
                    def method(self):
                        return "method"
                
                class InternalClass:
                    '''Not exported, should be flagged'''
                    pass
            """)
            )

            test_py = project_path / "test_example.py"
            test_py.write_text(
                dedent("""
                import unittest
                from main import used_function
                
                class TestExample(unittest.TestCase):
                    def test_used_function(self):
                        result = used_function(3)
                        self.assertEqual(result, 6)
                    
                    def test_unused_method(self):
                        pass
            """)
            )

            yield project_path

    def test_basic_analysis(self, temp_project):
        result_json = skylos.analyze(
            str(temp_project), exclude_folders=list(DEFAULT_EXCLUDE_FOLDERS)
        )
        result = json.loads(result_json)
        print("\n=== DEBUG ===")
        print("Unused functions:", [f["name"] for f in result["unused_functions"]])
        print("Unused imports:", [f["name"] for f in result["unused_imports"]])
        print("Whitelisted:", result.get("whitelisted", []))
        print("All definitions:", list(result.get("definitions", {}).keys())[:20])
        print("END OF DEBUG=============\n")

        assert "unused_functions" in result
        assert "unused_imports" in result
        assert "unused_variables" in result
        assert "unused_parameters" in result
        assert "unused_classes" in result

        assert len(result["unused_functions"]) > 0
        assert len(result["unused_imports"]) > 0

        unused_function_names = [f["name"] for f in result["unused_functions"]]
        assert any("unused_function" in name for name in unused_function_names), (
            f"Expected unused_function in {unused_function_names}"
        )

        assert "used_function" not in unused_function_names
        assert "main" not in unused_function_names

    def test_unused_imports_detection(self, temp_project):
        result_json = skylos.analyze(str(temp_project))
        result = json.loads(result_json)

        unused_imports = result["unused_imports"]
        import_names = [imp["name"] for imp in unused_imports]

        assert any("os" in name for name in import_names)
        assert any("typing.List" in name or "List" in name for name in import_names)

        used_imports = ["sys", "defaultdict"]
        for used_import in used_imports:
            assert not any(used_import in name for name in import_names)

    def test_class_detection(self, temp_project):
        result_json = skylos.analyze(str(temp_project))
        result = json.loads(result_json)

        unused_classes = result["unused_classes"]
        class_names = [cls["name"] for cls in unused_classes]

        assert any("UnusedClass" in name for name in class_names), (
            f"UnusedClass not found in {class_names}"
        )

        used_classes_flagged = [name for name in class_names if "UsedClass" in name]
        assert len(used_classes_flagged) == 0, (
            f"UsedClass was incorrectly flagged as unused: {used_classes_flagged}"
        )

        public_classes_flagged = [name for name in class_names if "PublicClass" in name]
        assert len(public_classes_flagged) == 0, (
            f"PublicClass was incorrectly flagged as unused: {public_classes_flagged}"
        )

    def test_exclude_folders(self, temp_project):
        result1 = skylos.analyze(str(temp_project))

        result2 = skylos.analyze(str(temp_project), exclude_folders=["mypackage"])

        data1 = json.loads(result1)
        data2 = json.loads(result2)

        total1 = sum(len(items) for items in data1.values() if isinstance(items, list))
        total2 = sum(len(items) for items in data2.values() if isinstance(items, list))

        assert total2 <= total1

    def test_custom_exclude_folders(self, temp_project):
        custom_dir = temp_project / "custom_exclude"
        custom_dir.mkdir()

        custom_file = custom_dir / "custom.py"
        custom_file.write_text(
            dedent("""
            def custom_function():
                return "should be excluded"
        """)
        )

        result_json = skylos.analyze(
            str(temp_project), exclude_folders=["custom_exclude"]
        )
        result = json.loads(result_json)

        all_files = []
        for category in [
            "unused_functions",
            "unused_imports",
            "unused_classes",
            "unused_variables",
        ]:
            for item in result[category]:
                all_files.append(item["file"])

        excluded_files = [f for f in all_files if "custom_exclude" in f]
        assert len(excluded_files) == 0, f"Found excluded files: {excluded_files}"

    def test_standalone_dirs_suppress_reachability_only(self, temp_project):
        tests_dir = temp_project / "tests"
        tests_dir.mkdir()
        (tests_dir / "test_api.py").write_text(
            dedent("""
            import math

            def helper():
                return 1

            class HelperSuite:
                pass
        """)
        )

        examples_dir = temp_project / "examples"
        examples_dir.mkdir()
        (examples_dir / "demo.py").write_text(
            dedent("""
            def example_main():
                return "demo"
        """)
        )

        benchmarks_dir = temp_project / "benchmarks"
        benchmarks_dir.mkdir()
        (benchmarks_dir / "bench_api.py").write_text(
            dedent("""
            class BenchmarkSuite:
                pass

            def time_parse():
                return 1
        """)
        )

        (temp_project / "test_example.py").write_text(
            dedent("""
            def root_test_helper():
                return 1
        """)
        )

        (temp_project / "conftest.py").write_text(
            dedent("""
            def root_conftest_helper():
                return 1
        """)
        )

        result = json.loads(skylos.analyze(str(temp_project)))

        unused_functions = {item["name"] for item in result["unused_functions"]}
        unused_classes = {item["name"] for item in result["unused_classes"]}
        unused_imports = {item["name"] for item in result["unused_imports"]}

        assert "helper" not in unused_functions
        assert "example_main" not in unused_functions
        assert "time_parse" not in unused_functions
        assert "root_test_helper" not in unused_functions
        assert "root_conftest_helper" not in unused_functions
        assert "HelperSuite" not in unused_classes
        assert "BenchmarkSuite" not in unused_classes
        assert "math" not in unused_imports

        suppressed_reasons = {
            item["reason"]
            for item in result.get("whitelisted", [])
            if "tests" in Path(item["file"]).parts
        }
        assert "test-only path" in suppressed_reasons

    def test_transitive_export_propagation(self, temp_project):
        pkg = temp_project / "pkg"
        pkg.mkdir()
        (pkg / "__init__.py").write_text("from pkg.events import Events\n")
        (pkg / "events.py").write_text(
            dedent("""
            class EventHook:
                def add_listener(self): pass
                def remove_listener(self): pass
                def _internal(self): pass

            class Events:
                request: EventHook
                response: EventHook
        """)
        )
        (temp_project / "app.py").write_text(
            dedent("""
            from pkg import Events

            e = Events()
            e.request.add_listener()
        """)
        )

        result = json.loads(skylos.analyze(str(temp_project)))
        unused_functions = {item["name"] for item in result["unused_functions"]}
        unused_classes = {item["name"] for item in result["unused_classes"]}

        assert "EventHook" not in unused_classes
        assert "remove_listener" not in unused_functions
        assert "add_listener" not in unused_functions
        assert any("_internal" in name for name in unused_functions)

    def test_magic_methods_excluded(self, temp_project):
        magic_py = temp_project / "magic.py"
        magic_py.write_text(
            dedent("""
            class MagicClass:
                def __init__(self):
                    self.value = 0
                
                def __str__(self):
                    return str(self.value)
                
                def __len__(self):
                    return self.value
                
                def regular_method(self):
                    return "regular"
        """)
        )

        result_json = skylos.analyze(str(temp_project))
        result = json.loads(result_json)

        unused_functions = result["unused_functions"]
        function_names = [f["name"] for f in unused_functions]

        magic_methods = ["__init__", "__str__", "__len__"]
        for magic_method in magic_methods:
            assert not any(magic_method in name for name in function_names)

    def test_test_methods_excluded(self, temp_project):
        """Test that test methods are not flagged as unused"""
        result_json = skylos.analyze(str(temp_project))
        result = json.loads(result_json)

        unused_functions = result["unused_functions"]
        function_names = [f["name"] for f in unused_functions]

        test_methods = ["test_used_function", "test_unused_method"]
        for test_method in test_methods:
            assert not any(test_method in name for name in function_names)

    def test_exported_functions_excluded(self, temp_project):
        result_json = skylos.analyze(str(temp_project))
        result = json.loads(result_json)

        unused_functions = result["unused_functions"]
        function_names = [f["name"] for f in unused_functions]

        assert not any("public_function" in name for name in function_names)

    def test_single_file_analysis(self, temp_project):
        main_file = temp_project / "main.py"

        result_json = skylos.analyze(str(main_file))
        result = json.loads(result_json)

        assert len(result["unused_functions"]) > 0
        assert len(result["unused_imports"]) > 0

        all_files = set()
        for category in ["unused_functions", "unused_imports", "unused_variables"]:
            for item in result[category]:
                all_files.add(Path(item["file"]).name)

        assert "main.py" in all_files
        assert "utils.py" not in all_files

    def test_confidence_levels(self, temp_project):
        result_json = skylos.analyze(str(temp_project))
        result = json.loads(result_json)

        for category in ["unused_functions", "unused_imports", "unused_variables"]:
            for item in result[category]:
                assert "confidence" in item
                assert isinstance(item["confidence"], (int, float))
                assert 0 <= item["confidence"] <= 100

    def test_analyzer_class_direct(self, temp_project):
        analyzer = Skylos()
        result_json = analyzer.analyze(str(temp_project), thr=50)
        result = json.loads(result_json)

        expected_keys = [
            "unused_functions",
            "unused_imports",
            "unused_classes",
            "unused_variables",
            "unused_parameters",
            "analysis_summary",
        ]

        for key in expected_keys:
            assert key in result

    def test_empty_project(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            result_json = skylos.analyze(temp_dir)
            result = json.loads(result_json)

            assert result["unused_functions"] == []
            assert result["unused_imports"] == []
            assert result["analysis_summary"]["total_files"] == 0

    def test_threshold_filtering(self, temp_project):
        result_high = json.loads(skylos.analyze(str(temp_project), conf=95))

        result_low = json.loads(skylos.analyze(str(temp_project), conf=10))

        assert len(result_high["unused_functions"]) <= len(
            result_low["unused_functions"]
        )
        assert len(result_high["unused_imports"]) <= len(result_low["unused_imports"])


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



"""Skylos AI Defense Engine — checks LLM integrations for missing guardrails."""

from skylos.defend.result import DefenseResult, DefenseScore, OpsScore
from skylos.defend.engine import run_defense_checks
from skylos.defend.scoring import (
    compute_defense_score,
    compute_ops_score,
    SEVERITY_WEIGHTS,
)
from skylos.defend.report import format_defense_table, format_defense_json

__all__ = [
    "DefenseResult",
    "DefenseScore",
    "OpsScore",
    "run_defense_checks",
    "compute_defense_score",
    "compute_ops_score",
    "SEVERITY_WEIGHTS",
    "format_defense_table",
    "format_defense_json",
]



from __future__ import annotations

import libcst as cst
from libcst.helpers import get_full_name_for_node
from libcst.metadata import PositionProvider


class _CommentOutBlock(cst.CSTTransformer):
    METADATA_DEPENDENCIES = (PositionProvider,)

    def __init__(self, module_code: str, marker: str = "SKYLOS DEADCODE"):
        self.module_code = module_code.splitlines(True)
        self.marker = marker

    def _comment_block(self, start_line, end_line):
        lines = self.module_code[start_line - 1 : end_line]
        out: list[cst.EmptyLine] = []

        out.append(
            cst.EmptyLine(
                comment=cst.Comment(
                    f"# {self.marker} START (lines {start_line}-{end_line})"
                )
            )
        )

        for raw in lines:
            out.append(cst.EmptyLine(comment=cst.Comment("# " + raw.rstrip("\n"))))

        out.append(cst.EmptyLine(comment=cst.Comment(f"# {self.marker} END")))
        return out

    def _replace_statement_with_comment_block(
        self, orig_stmt: cst.SimpleStatementLine
    ) -> cst.SimpleStatementLine:
        pos = self.get_metadata(PositionProvider, orig_stmt)
        leading = self._comment_block(pos.start.line, pos.end.line)

        return cst.SimpleStatementLine(
            body=[cst.Pass()],
            leading_lines=leading,
        )


class _CommentOutFunctionAtLine(_CommentOutBlock):
    def __init__(self, func_name, target_line, module_code, marker):
        super().__init__(module_code, marker)
        self.func_name = func_name
        self.target_line = target_line
        self.changed = False

    def _is_target(self, node: cst.CSTNode):
        pos = self.get_metadata(PositionProvider, node, None)
        return bool(pos and pos.start.line == self.target_line)

    def leave_FunctionDef(self, orig: cst.FunctionDef, updated: cst.FunctionDef):
        target = self.func_name.split(".")[-1]
        if self._is_target(orig) and orig.name.value == target:
            self.changed = True
            pos = self.get_metadata(PositionProvider, orig)
            leading = self._comment_block(pos.start.line, pos.end.line)
            return cst.SimpleStatementLine(body=[cst.Pass()], leading_lines=leading)
        return updated

    def leave_AsyncFunctionDef(
        self, orig: cst.AsyncFunctionDef, updated: cst.AsyncFunctionDef
    ):
        target = self.func_name.split(".")[-1]
        if self._is_target(orig) and orig.name.value == target:
            self.changed = True
            pos = self.get_metadata(PositionProvider, orig)
            leading = self._comment_block(pos.start.line, pos.end.line)
            return cst.SimpleStatementLine(body=[cst.Pass()], leading_lines=leading)
        return updated


class _CommentOutImportAtLine(_CommentOutBlock):
    def __init__(self, target_name, target_line, module_code, marker):
        super().__init__(module_code, marker)
        self.target_name = target_name
        self.target_line = target_line
        self.changed = False

    def _is_target_line(self, node: cst.CSTNode):
        pos = self.get_metadata(PositionProvider, node, None)
        return bool(pos and (pos.start.line <= self.target_line <= pos.end.line))

    def _name_code(self, node: cst.CSTNode):
        code = get_full_name_for_node(node) or ""
        if code:
            return code
        if hasattr(node, "value"):
            v = getattr(node, "value")
            if isinstance(v, str):
                return v
        return ""

    def _render_single_alias_text(self, head, alias: cst.ImportAlias, is_from):
        name_code = self._name_code(alias.name)
        alias_txt = name_code
        if alias.asname:
            alias_txt = f"{alias_txt} as {alias.asname.name.value}"

        if is_from:
            return f"from {head} import {alias_txt}"
        return f"import {alias_txt}"

    def _split_aliases(self, aliases, head, is_from):
        kept = []
        removed_for_comment: list[str] = []

        for alias in list(aliases):
            bound = _bound_name_for_import_alias(alias)

            if (not is_from) and _is_plain_dotted_import(alias):
                match_keys = {bound}
            else:
                name_code = self._name_code(alias.name)
                tail = name_code.split(".")[-1] if name_code else ""
                if tail:
                    match_keys = {bound, tail}
                else:
                    match_keys = {bound}

            if self.target_name in match_keys:
                self.changed = True
                removed_for_comment.append(
                    self._render_single_alias_text(head, alias, is_from)
                )
            else:
                kept.append(alias)

        return kept, removed_for_comment

    def _add_leading_comment_line(
        self, line: cst.SimpleStatementLine, text
    ) -> cst.SimpleStatementLine:
        comment = cst.EmptyLine(comment=cst.Comment(f"# {text}"))
        leading = list(line.leading_lines)
        leading.append(comment)
        return line.with_changes(leading_lines=leading)

    def leave_SimpleStatementLine(
        self, orig: cst.SimpleStatementLine, updated: cst.SimpleStatementLine
    ):
        if not self._is_target_line(orig):
            return updated

        if len(updated.body) != 1:
            return updated

        stmt = updated.body[0]

        if isinstance(stmt, cst.Import):
            kept, removed = self._split_aliases(stmt.names, head="", is_from=False)
            if not removed:
                return updated

            if not kept:
                return self._replace_statement_with_comment_block(orig)

            new_import = stmt.with_changes(names=tuple(kept))
            new_line = updated.with_changes(body=[new_import])

            removed_txt = "; ".join(removed)
            return self._add_leading_comment_line(
                new_line, f"{self.marker}: {removed_txt}"
            )

        if isinstance(stmt, cst.ImportFrom):
            if isinstance(stmt.names, cst.ImportStar):
                return updated

            if stmt.relative:
                dots = "." * len(stmt.relative)
            else:
                dots = ""

            modname = ""
            if stmt.module is not None:
                modname = self._name_code(stmt.module)

            head = f"{dots}{modname}"

            kept, removed = self._split_aliases(list(stmt.names), head, is_from=True)
            if not removed:
                return updated

            if not kept:
                return self._replace_statement_with_comment_block(orig)

            new_from = stmt.with_changes(names=tuple(kept))
            new_line = updated.with_changes(body=[new_from])

            removed_txt = "; ".join(removed)
            return self._add_leading_comment_line(
                new_line, f"{self.marker}: {removed_txt}"
            )

        return updated


def comment_out_unused_function_cst(
    code, func_name, line_number, marker="SKYLOS DEADCODE"
):
    wrapper = cst.MetadataWrapper(cst.parse_module(code))
    tx = _CommentOutFunctionAtLine(func_name, line_number, code, marker)
    new_mod = wrapper.visit(tx)
    return new_mod.code, tx.changed


def comment_out_unused_import_cst(
    code, import_name, line_number, marker="SKYLOS DEADCODE"
):
    wrapper = cst.MetadataWrapper(cst.parse_module(code))
    tx = _CommentOutImportAtLine(import_name, line_number, code, marker)
    new_mod = wrapper.visit(tx)
    return new_mod.code, tx.changed


def _bound_name_for_import_alias(alias: cst.ImportAlias):
    if alias.asname:
        return alias.asname.name.value
    node = alias.name
    while isinstance(node, cst.Attribute):
        node = node.value
    return node.value


def _is_plain_dotted_import(alias: cst.ImportAlias):
    return (alias.asname is None) and isinstance(alias.name, cst.Attribute)


class _RemoveImportAtLine(cst.CSTTransformer):
    METADATA_DEPENDENCIES = (PositionProvider,)

    def __init__(self, target_name: str, target_line: int):
        self.target_name = target_name
        self.target_line = target_line
        self.changed = False

    def _is_target_line(self, node: cst.CSTNode):
        pos = self.get_metadata(PositionProvider, node, None)
        return bool(pos and (pos.start.line <= self.target_line <= pos.end.line))

    def _filter_aliases(self, aliases):
        kept = []
        for alias in aliases:
            bound = _bound_name_for_import_alias(alias)

            if _is_plain_dotted_import(alias):
                match_keys = {bound}
            else:
                name_code = get_full_name_for_node(alias.name) or ""
                tail = name_code.split(".")[-1] if name_code else ""
                if tail:
                    match_keys = {bound, tail}
                else:
                    match_keys = {bound}

            if self.target_name in match_keys:
                self.changed = True
                continue

            kept.append(alias)
        return kept

    def leave_Import(self, orig: cst.Import, updated: cst.Import):
        if not self._is_target_line(orig):
            return updated
        kept = self._filter_aliases(updated.names)
        if not kept:
            return cst.RemoveFromParent()
        return updated.with_changes(names=tuple(kept))

    def leave_ImportFrom(self, orig: cst.ImportFrom, updated: cst.ImportFrom):
        if not self._is_target_line(orig):
            return updated
        if isinstance(updated.names, cst.ImportStar):
            return updated
        kept = self._filter_aliases(list(updated.names))
        if not kept:
            return cst.RemoveFromParent()
        return updated.with_changes(names=tuple(kept))


class _RemoveFunctionAtLine(cst.CSTTransformer):
    METADATA_DEPENDENCIES = (PositionProvider,)

    def __init__(self, func_name, target_line):
        self.func_name = func_name
        self.target_line = target_line
        self.changed = False

    def _is_target(self, node: cst.CSTNode):
        pos = self.get_metadata(PositionProvider, node, None)
        return bool(pos and pos.start.line == self.target_line)

    def leave_FunctionDef(self, orig: cst.FunctionDef, updated: cst.FunctionDef):
        target = self.func_name.split(".")[-1]
        if self._is_target(orig) and (orig.name.value == target):
            self.changed = True
            return cst.RemoveFromParent()
        return updated

    def leave_AsyncFunctionDef(
        self, orig: cst.AsyncFunctionDef, updated: cst.AsyncFunctionDef
    ):
        target = self.func_name.split(".")[-1]
        if self._is_target(orig) and (orig.name.value == target):
            self.changed = True
            return cst.RemoveFromParent()
        return updated


def remove_unused_import_cst(code, import_name, line_number):
    wrapper = cst.MetadataWrapper(cst.parse_module(code))
    tx = _RemoveImportAtLine(import_name, line_number)
    new_mod = wrapper.visit(tx)
    return new_mod.code, tx.changed


def remove_unused_function_cst(code, func_name, line_number):
    wrapper = cst.MetadataWrapper(cst.parse_module(code))
    tx = _RemoveFunctionAtLine(func_name, line_number)
    new_mod = wrapper.visit(tx)
    return new_mod.code, tx.changed


class _RemoveClassAtLine(cst.CSTTransformer):
    METADATA_DEPENDENCIES = (PositionProvider,)

    def __init__(self, class_name, target_line):
        self.class_name = class_name
        self.target_line = target_line
        self.changed = False

    def _is_target(self, node: cst.CSTNode):
        pos = self.get_metadata(PositionProvider, node, None)
        return bool(pos and pos.start.line == self.target_line)

    def leave_ClassDef(self, orig: cst.ClassDef, updated: cst.ClassDef):
        target = self.class_name.split(".")[-1]
        if self._is_target(orig) and (orig.name.value == target):
            self.changed = True
            return cst.RemoveFromParent()
        return updated


class _RemoveVariableAtLine(cst.CSTTransformer):
    METADATA_DEPENDENCIES = (PositionProvider,)

    def __init__(self, var_name, target_line):
        self.var_name = var_name
        self.target_line = target_line
        self.changed = False

    def _is_target_line(self, node: cst.CSTNode):
        pos = self.get_metadata(PositionProvider, node, None)
        return bool(pos and (pos.start.line <= self.target_line <= pos.end.line))

    def leave_SimpleStatementLine(
        self, orig: cst.SimpleStatementLine, updated: cst.SimpleStatementLine
    ):
        if not self._is_target_line(orig):
            return updated
        if len(updated.body) != 1:
            return updated
        stmt = updated.body[0]
        if isinstance(stmt, (cst.Assign, cst.AnnAssign)):
            self.changed = True
            return cst.RemoveFromParent()
        return updated


def remove_unused_class_cst(code, class_name, line_number):
    """Remove an unused class definition at the given line."""
    wrapper = cst.MetadataWrapper(cst.parse_module(code))
    tx = _RemoveClassAtLine(class_name, line_number)
    new_mod = wrapper.visit(tx)
    return new_mod.code, tx.changed


def remove_unused_variable_cst(code, var_name, line_number):
    """Remove an unused variable assignment at the given line."""
    wrapper = cst.MetadataWrapper(cst.parse_module(code))
    tx = _RemoveVariableAtLine(var_name, line_number)
    new_mod = wrapper.visit(tx)
    return new_mod.code, tx.changed



import json
import os
import shutil
import subprocess
from pathlib import Path

import skylos
from skylos.constants import DEFAULT_EXCLUDE_FOLDERS
from .go_contract import build_go_engine_args, validate_go_engine_output


DEFAULT_SKIP_DIRS = {d for d in DEFAULT_EXCLUDE_FOLDERS if "*" not in d}


class GoEngineError(RuntimeError):
    pass


def discover_go_modules(scan_root):
    scan_root = Path(scan_root).resolve()
    modules = []

    if (scan_root / "go.mod").is_file():
        return [scan_root]

    stack = [scan_root]
    while stack:
        cur = stack.pop()

        name = cur.name
        if name in DEFAULT_SKIP_DIRS:
            continue
        if name.startswith(".") and name not in {".", ".."} and name != ".github":
            continue

        go_mod = cur / "go.mod"
        if go_mod.is_file():
            modules.append(cur)
            continue

        try:
            for child in cur.iterdir():
                if child.is_dir():
                    stack.append(child)
        except Exception:
            continue

    modules = sorted(modules, key=lambda p: str(p))
    return modules


def resolve_go_engine_bin():
    override = os.getenv("SKYLOS_GO_BIN")
    if override:
        p = Path(override).expanduser()
        if p.is_file():
            return str(p)
        raise GoEngineError("SKYLOS_GO_BIN is set but binary does not exist: %s" % p)

    exe = "skylos-go.exe" if os.name == "nt" else "skylos-go"
    found = shutil.which(exe) or shutil.which("skylos-go")
    if found:
        return found

    pkg_dir = Path(__file__).resolve().parent / "go" / exe
    if pkg_dir.is_file():
        return str(pkg_dir)

    raise GoEngineError(
        "Go engine binary not found (skylos-go).\n"
        "Build it locally:\n"
        "  cd engines/go && go build -o skylos-go ./cmd/skylos-go\n"
        "Then set:\n"
        "  export SKYLOS_GO_BIN=/absolute/path/to/skylos-go"
    )


def run_go_engine_for_module(module_root, timeout_s=60):
    engine_bin = resolve_go_engine_bin()
    module_root = Path(module_root).resolve()

    argv = build_go_engine_args(
        engine_bin=engine_bin,
        root=str(module_root),
        skylos_version=str(skylos.__version__),
    )

    try:
        proc = subprocess.run(
            argv,
            cwd=str(module_root),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            timeout=timeout_s,
            check=False,
        )
    except subprocess.TimeoutExpired:
        raise GoEngineError(
            "Go engine timed out after %ss for module: %s" % (timeout_s, module_root)
        )
    except Exception as e:
        raise GoEngineError("Failed to run Go engine: %s" % e)

    if proc.returncode != 0:
        raise GoEngineError(
            "Go engine failed.\n"
            "Command: %s\n"
            "Exit code: %s\n"
            "STDERR:\n%s"
            % (" ".join(argv), proc.returncode, (proc.stderr or "").strip())
        )

    try:
        obj = json.loads(proc.stdout or "")
    except Exception as e:
        raise GoEngineError(
            "Go engine returned invalid JSON.\n"
            "STDOUT:\n%s\n"
            "STDERR:\n%s\n"
            "Error: %s" % ((proc.stdout or "").strip(), (proc.stderr or "").strip(), e)
        )

    out = validate_go_engine_output(obj)
    return {
        "findings": list(out.get("findings", [])),
        "symbols": out.get("symbols"),
    }



from skylos.defend.plugin import DefensePlugin
from skylos.defend.result import DefenseResult
from skylos.discover.integration import LLMIntegration
from skylos.discover.graph import AIIntegrationGraph


class InputLengthLimitPlugin(DefensePlugin):
    id = "input-length-limit"
    name = "Input Length Limited"
    severity = "low"
    owasp_llm = "LLM01"
    description = (
        "User input should have a length/size check before reaching the LLM call"
    )
    remediation = (
        "Add input length validation (e.g. len(input) check, [:N] slicing, "
        "or max_length validator) before the LLM call."
    )

    def applies_to(self, integration: LLMIntegration) -> bool:
        return bool(integration.input_sources)

    def check(
        self, integration: LLMIntegration, graph: AIIntegrationGraph
    ) -> DefenseResult:
        if integration.has_input_length_limit:
            loc = integration.input_length_limit_location or integration.location
            return self._pass(
                integration,
                loc,
                f"Input length check at {loc}",
            )

        return self._fail(
            integration,
            integration.location,
            "Unbounded user input reaches LLM call",
        )



import ast
from pathlib import Path
from skylos.rules.base import SkylosRule

RULE_ID = "SKY-Q301"
COGNITIVE_RULE_ID = "SKY-Q306"

_COMPLEX_NODES = (
    ast.If,
    ast.For,
    ast.AsyncFor,
    ast.While,
    ast.IfExp,
)


def _func_complexity(fn_node: ast.AST) -> int:
    c = 1

    class Visitor(ast.NodeVisitor):
        def visit_FunctionDef(self, _):
            return

        def visit_AsyncFunctionDef(self, _):
            return

        def visit_ClassDef(self, _):
            return

        def visit_Lambda(self, _):
            return

        def generic_visit(self, node):
            nonlocal c

            if isinstance(node, _COMPLEX_NODES):
                c += 1

            if isinstance(node, ast.BoolOp):
                c += max(len(node.values) - 1, 0)

            if isinstance(node, ast.Try):
                c += len(getattr(node, "handlers", []) or [])

            if hasattr(ast, "TryStar") and isinstance(node, ast.TryStar):
                c += len(getattr(node, "handlers", []) or [])

            if hasattr(ast, "Match") and isinstance(node, ast.Match):
                cases = getattr(node, "cases", []) or []
                c += max(len(cases) - 1, 0)
                for case in cases:
                    if getattr(case, "guard", None) is not None:
                        c += 1

            if isinstance(node, ast.comprehension):
                c += 1
                c += len(node.ifs)

            for child in ast.iter_child_nodes(node):
                self.visit(child)

    v = Visitor()
    for stmt in fn_node.body:
        v.visit(stmt)
    return c


def _func_length(node):
    start = getattr(node, "lineno", None)
    end = getattr(node, "end_lineno", None)

    if start is None:
        return 0

    if end is None:
        end = start
        for child in ast.walk(node):
            ln = getattr(child, "lineno", None)
            if ln is not None and ln > end:
                end = ln

    return max(end - start + 1, 0)


class ComplexityRule(SkylosRule):
    rule_id = "SKY-Q301"
    name = "Cyclomatic Complexity"

    def __init__(self, threshold=10):
        self.threshold = threshold

    def visit_node(self, node, context):
        if not isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
            return None

        complexity = _func_complexity(node)

        if complexity <= self.threshold:
            return None

        if complexity < 15:
            severity = "WARN"
        elif complexity < 25:
            severity = "HIGH"
        else:
            severity = "CRITICAL"

        length = _func_length(node)
        mod = context.get("mod", "")

        if mod:
            func_name = f"{mod}.{node.name}"
        else:
            func_name = node.name

        return [
            {
                "rule_id": self.rule_id,
                "kind": "complexity",
                "severity": severity,
                "type": "function",
                "name": func_name,
                "simple_name": node.name,
                "value": complexity,
                "threshold": self.threshold,
                "length": length,
                "message": f"Cyclomatic complexity is {complexity} (threshold: {self.threshold}). Consider splitting branches.",
                "file": context.get("filename"),
                "basename": Path(context.get("filename", "")).name,
                "line": node.lineno,
            }
        ]


def _cognitive_complexity(fn_node: ast.AST) -> int:

    total = 0

    def _process(node, nesting: int, is_elif: bool = False):
        nonlocal total

        if isinstance(
            node, (ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef, ast.Lambda)
        ):
            for child in ast.iter_child_nodes(node):
                _process(child, 0)
            return

        if isinstance(node, ast.If):
            if is_elif:
                total += 1
            else:
                total += 1 + nesting

            inner_nesting = nesting + 1

            _process(node.test, inner_nesting)

            for body_child in node.body:
                _process(body_child, inner_nesting)

            if node.orelse:
                if len(node.orelse) == 1 and isinstance(node.orelse[0], ast.If):
                    _process(node.orelse[0], nesting, is_elif=True)
                else:
                    total += 1
                    for else_child in node.orelse:
                        _process(else_child, inner_nesting)
            return

        if isinstance(node, (ast.For, ast.AsyncFor, ast.While)):
            total += 1 + nesting
            if isinstance(node, ast.While):
                _process(node.test, nesting + 1)
            for body_child in node.body:
                _process(body_child, nesting + 1)
            if node.orelse:
                for else_child in node.orelse:
                    _process(else_child, nesting + 1)
            return

        if isinstance(node, ast.BoolOp):
            total += max(len(node.values) - 1, 0)
            for val in node.values:
                _process(val, nesting)
            return

        if isinstance(node, ast.IfExp):
            total += 1 + nesting
            _process(node.test, nesting)
            _process(node.body, nesting)
            _process(node.orelse, nesting)
            return

        if isinstance(node, ast.Try):
            for handler in getattr(node, "handlers", []) or []:
                total += 1 + nesting
                for h_child in handler.body:
                    _process(h_child, nesting + 1)
            for body_child in getattr(node, "body", []) or []:
                _process(body_child, nesting)
            for fb_child in getattr(node, "finalbody", []) or []:
                _process(fb_child, nesting)
            for else_child in getattr(node, "orelse", []) or []:
                _process(else_child, nesting)
            return

        if hasattr(ast, "TryStar") and isinstance(node, ast.TryStar):
            for handler in getattr(node, "handlers", []) or []:
                total += 1 + nesting
                for h_child in handler.body:
                    _process(h_child, nesting + 1)
            for body_child in getattr(node, "body", []) or []:
                _process(body_child, nesting)
            for fb_child in getattr(node, "finalbody", []) or []:
                _process(fb_child, nesting)
            for else_child in getattr(node, "orelse", []) or []:
                _process(else_child, nesting)
            return

        if hasattr(ast, "Match") and isinstance(node, ast.Match):
            total += 1 + nesting
            for case in getattr(node, "cases", []) or []:
                for case_child in case.body:
                    _process(case_child, nesting + 1)
            return

        for child in ast.iter_child_nodes(node):
            _process(child, nesting)

    for stmt in fn_node.body:
        _process(stmt, 0)

    return total


class CognitiveComplexityRule(SkylosRule):
    rule_id = COGNITIVE_RULE_ID
    name = "Cognitive Complexity"

    def __init__(self, threshold=15):
        self.threshold = threshold

    def visit_node(self, node, context):
        if not isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
            return None

        complexity = _cognitive_complexity(node)

        if complexity <= self.threshold:
            return None

        if complexity <= 30:
            severity = "MEDIUM"
        elif complexity <= 50:
            severity = "HIGH"
        else:
            severity = "CRITICAL"

        mod = context.get("mod", "")
        if mod:
            func_name = f"{mod}.{node.name}"
        else:
            func_name = node.name

        return [
            {
                "rule_id": self.rule_id,
                "kind": "complexity",
                "severity": severity,
                "type": "function",
                "name": func_name,
                "simple_name": node.name,
                "value": complexity,
                "threshold": self.threshold,
                "message": f"Cognitive complexity is {complexity} (threshold: {self.threshold}). Consider simplifying nested logic.",
                "file": context.get("filename"),
                "basename": Path(context.get("filename", "")).name,
                "line": node.lineno,
            }
        ]



from __future__ import annotations

from typing import Any

DEAD_CODE_RESULT_KEYS: tuple[str, ...] = (
    "unused_functions",
    "unused_classes",
    "unused_variables",
    "unused_imports",
    "unused_parameters",
)


def collect_dead_code_findings(result: dict[str, Any]) -> list[dict[str, Any]]:
    findings: list[dict[str, Any]] = []
    for key in DEAD_CODE_RESULT_KEYS:
        for finding in result.get(key, []) or []:
            findings.append(dict(finding))
    return findings



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional


@dataclass
class DefenseResult:
    plugin_id: str
    passed: bool
    integration_location: str
    location: str
    message: str
    severity: str  # "critical" | "high" | "medium" | "low"
    weight: int
    category: str  # "defense" or "ops"
    owasp_llm: Optional[str] = None
    remediation: str = ""

    def to_dict(self) -> dict:
        return {
            "plugin_id": self.plugin_id,
            "passed": self.passed,
            "integration_location": self.integration_location,
            "location": self.location,
            "message": self.message,
            "severity": self.severity,
            "weight": self.weight,
            "category": self.category,
            "owasp_llm": self.owasp_llm,
            "remediation": self.remediation,
        }


@dataclass
class DefenseScore:
    weighted_score: int
    weighted_max: int
    score_pct: int
    risk_rating: str
    passed: int
    total: int

    def to_dict(self) -> dict:
        return {
            "weighted_score": self.weighted_score,
            "weighted_max": self.weighted_max,
            "score_pct": self.score_pct,
            "risk_rating": self.risk_rating,
            "passed": self.passed,
            "total": self.total,
        }


@dataclass
class OpsScore:
    passed: int
    total: int
    score_pct: int
    rating: str  # "EXCELLENT" | "GOOD" | "FAIR" | "POOR"

    def to_dict(self) -> dict:
        return {
            "passed": self.passed,
            "total": self.total,
            "score_pct": self.score_pct,
            "rating": self.rating,
        }
